/*
 * Implementation of arrays
 */
#include "s.h"
#include "heap.h"


void scm_array_mark(SOBJ arr)
{
  int i;
  SOBJ *slot;

  slot = SCM_ARRAY(arr);
  if (slot == NULL) 	return;
  for (i = 0; i < SCM_ASIZE(arr); i++) {
	if (!SCM_INUMP(slot[i]) && scm_is_pointer_to_heap(slot[i])) {
	  scm_gc_mark(slot[i]);
	}	
  }
}

void scm_array_sweep(SOBJ arr)
{
  SCM_Array *p = SCM_ADESCR(arr);

#ifdef DEBUG
  scm_puts("; removing array #");  port_putx(scm_outp, arr);
  scm_puts("\n");
#endif

  if (p != NULL) {
	scm_free(p);
  }
  SCM_ADESCR(arr) = NULL;
}

/*-- array reconizer and parser */
int scm_array_reconize(PORT *port, int c)
{
  int next = port_peekc(port);
  return (c == '#' && next == '(');
}

SOBJ scm_array_parse(PORT *port, int start_char)
{
  port_getc(port);			/* ignore the leading '(' */
  return(scm_list_to_vector(scm_read_list(port)));
}



SOBJ scm_mkarray(int size, SOBJ fill)
{
  SOBJ arr, *p;
  int i;

  arr = scm_newcell(SOBJ_T_ARRAY);
  SCM_ADESCR(arr) =
	scm_must_alloc(sizeof(SCM_Array) + (size * sizeof(SOBJ)));
  
  SCM_ADESCR(arr)->size = size;
  SCM_ADESCR(arr)->alloced = size + 1;

  p = SCM_ARRAY(arr);
  for (i = 0; i < size; i++)  *p++ = fill;
  return(arr);
}

/************************************************************************
 * Vector type definition
 ************************************************************************/

void scm_vector_print(SOBJ arr, PORT *p)
{
  int i;
  port_puts(p, "#(");
  for (i = 0; i < SCM_ASIZE(arr); i++) {
	scm_write_obj(SCM_AREF(arr, i), p, 1);
	if (i < (SCM_ASIZE(arr) -1)) port_putc(p, ' ');
  }
  port_putc(p, ')');
}

void scm_vector_write(SOBJ arr, PORT *p)
{
  int i;
  port_puts(p, "#(");
  for (i = 0; i < SCM_ASIZE(arr); i++) {
	scm_write_obj(SCM_AREF(arr, i), p, 0);
	if (i < (SCM_ASIZE(arr) -1)) port_putc(p, ' ');
  }
  port_putc(p, ')');
}

/****************************************************************
 * R5RS procedures
 ****************************************************************/

SOBJ scm_vectorp(SOBJ obj)
{
  return( SCM_ARRAYP(obj) ? scm_true : scm_false );
}

SOBJ scm_make_vector(int n, SOBJ *p)
{
  SOBJ fill;
  if (n < 1 || !SCM_INUMP(p[0])) 	err("make-vector: bad count", p[0]);
  fill = (n >=2) ? p[1] : scm_unbound;
  return(scm_mkarray(SCM_INUM(p[0]), fill));
}

SOBJ scm_vector(int n, SOBJ *p)
{
  int i;
  SOBJ new = scm_mkarray(n, NULL);
  for (i = 0; i < n; i++) {
	SCM_AREF(new, i) = p[i];
  }
  return(new);
}

SOBJ scm_vector_length(SOBJ obj)
{
  return(SCM_MKINUM(SCM_ASIZE(obj)));
}

SOBJ scm_vector_ref(SOBJ vector, SOBJ index)
{
  int i;

  if (!SCM_ARRAYP(vector))	err("vector-ref: bad vector", vector);
  if (!SCM_INUMP(index))	err("vector-ref: bad index", index);
  i = SCM_INUM(index);
  if (i < 0 || i >= SCM_ASIZE(vector))
	err("vector-ref: out of range", index);

  return(SCM_AREF(vector, i));
}

SOBJ scm_vector_set(SOBJ vector, SOBJ index, SOBJ value)
{
  int i;

  if (!SCM_ARRAYP(vector))	err("vector-set!: bad vector", vector);
  if (!SCM_INUMP(index))	err("vector-set!: bad index", index);
  i = SCM_INUM(index);
  if (i < 0 || i >= SCM_ASIZE(vector))
	err("vector-set!: out of range", index);

  SCM_AREF(vector, i) = value;
  return(scm_undefined);
}

SOBJ scm_vector_to_list(SOBJ vector)
{
  int i;
  SOBJ list = NULL;

  if (!SCM_ARRAYP(vector)) 	err("vector->list: bad vector", vector);
  i = SCM_ASIZE(vector);
  while(--i >= 0) {
	list = scm_cons(SCM_AREF(vector, i), list);
  }
  return(list);
}

SOBJ scm_list_to_vector(SOBJ list)
{
  int i, len = scm_list_length(list);
  SOBJ v;
  
  if (len < 0)	err("list->vector: bad list", list);
  v = scm_mkarray(len, scm_unbound);
  i = 0;
  while(list) {
	SCM_AREF(v, i) = SCM_CAR(list);
	list = SCM_CDR(list);
	i++;
  }
  return(v);
}

SOBJ scm_vector_fill(SOBJ vector, SOBJ fill)
{
  int i;
  if (!SCM_ARRAYP(vector))	err("vector-fill! bad vector", vector);
  
  for (i = 0; i < SCM_ASIZE(vector); i++) {
	SCM_AREF(vector, i) = fill;
  }
  return(scm_undefined);
}

/****************************************************************
 * Extensions
 ****************************************************************/

SOBJ scm_vector_resize(SOBJ vector, SOBJ size)
{
  int newsize;
  
  if (!SCM_ARRAYP(vector))	err("vector-resize: bad vector", vector);
  if (!SCM_INUMP(size))		err("vector-resize: bad size",   size);

  newsize = SCM_INUM(size);
  if (newsize > SCM_ADESCR(vector)->alloced) {
	SCM_Array *p;
	int i;

	p = scm_must_realloc(SCM_ADESCR(vector),
						 sizeof(SCM_Array) + (newsize * sizeof(SOBJ)));

	for (i = p->size; i < newsize; i++) {
	  p->item[i] = scm_unbound;
	}
	SCM_ADESCR(vector) = p;
	p->size = newsize;
	p->alloced = newsize + 1;
  } else {
	SCM_ASIZE(vector) = newsize;
  }
  return(vector);
}

SOBJ scm_vector_append(SOBJ vector, SOBJ item)
{
  int i = SCM_ASIZE(vector);
  scm_vector_resize(vector, SCM_MKINUM(i+1));
  SCM_AREF(vector, i) = item;
  return(vector);
}




/****************************************************************
 * Initialisation
 ****************************************************************/

void scm_init_array()
{
  /* R5RS */
  scm_add_cprim("vector?",			scm_vectorp,		1);
  scm_add_cprim("make-vector",		scm_make_vector,	-1);
  scm_add_cprim("vector",			scm_vector,			-1);
  scm_add_cprim("vector-length",	scm_vector_length,	1);
  scm_add_cprim("vector-ref",		scm_vector_ref,		2);
  scm_add_cprim("vector-set!",		scm_vector_set,		3);
  scm_add_cprim("vector->list",		scm_vector_to_list,	1);
  scm_add_cprim("list->vector",		scm_list_to_vector,	1);
  scm_add_cprim("vector-fill!",		scm_vector_fill,	2);

  /* extensions */
  scm_add_cprim("vector-resize",	scm_vector_resize,  2);
  scm_add_cprim("vector-append",	scm_vector_append,  2);
  
}
